% ************************************************************************
% author:   Amin Zheng                                                  **
% date:     06/23/2015                                                  **
% modified:                                                             **
% purpose:  arithmetic edge encoding with context tree                  **
% ************************************************************************

% input: direction code (1-4 means North, East, South, West),
% output: AEC codeword (d) and the number of used bits (bCount).

% reference: Amir Said, "Introduction to Arithmetic Coding - Theory and Practice", April 2004.
% http://www.hpl.hp.com/techreports/2004/HPL-2004-76.pdf

function [d, bCount, contextBitsOut] = aecMCT (maxOrder, dcode, countStateSpace, treeHeight, contextBits)

% for debug ***
contextBitsOut = contextBits;

% step 0: initialize var
b = 0;   % interval base
l = 1;   % interval length
bCount = 0;  
gamma = 2;  % interval rescaling factor
deltaOffset = 1/3;

prev = zeros(maxOrder,1);

stat2IndVec = zeros(maxOrder, maxOrder);
for i = 1:maxOrder
    stat2IndVec(i, 1:i) = 3.^(0:i-1);
end

tempCcode = zeros(length(dcode), 1);


for ind = 1:length(dcode),
     
    tempMaxOrder = 0;
    
    % step 1: find best estimate of prob of 3 poss. directions ********
    if ind == 1,
        prob = ones(1,4) / 4.0;    % first chain
    else
        if ind > 2,
            % get the largest context length ***
            tempMaxOrder = min(ind - 2, maxOrder);
            % get the previous DCC symbols ***
            prev(1:tempMaxOrder) = tempCcode(ind-tempMaxOrder:ind-1);
            
            % get the context length of current symbol ***
            heightIndex = stat2IndVec(tempMaxOrder, 1:tempMaxOrder)...
                * (prev(1:tempMaxOrder) - 1) + 1;
            tempCurOrder = treeHeight(tempMaxOrder, heightIndex);
            
            % get the prob according to context tree ***
            prohInd = stat2IndVec(tempCurOrder, 1:tempCurOrder)...
                * (prev(tempMaxOrder-tempCurOrder+1:tempMaxOrder) - 1) + 1;
            prob = countStateSpace(tempCurOrder, prohInd, :);
            prob = prob(:)';
            
            prob = (prob + deltaOffset) / (3*deltaOffset + sum(prob));
        else
            prob = ones(1,3) / 3.0;
        end  % if ind > minFitLen,
        
        % add end-code prob, can be any function of ind
        endProb = 0;   % here end-code prob unused, code the symbol number instead
        prob = [prob*(1.0-endProb)  endProb];
    end % if ind == 1

    % step 2: actual arithmetic coding:  given prob. vector, index, ********
    % compute new interval ********

    % get the whole turning chain to help train the probabilty (l,s,r as stats)
    if ind == 1,
        ccode = dcode(1);  % fist chain is in absolute term  
    elseif ind <= length(dcode),
        if dcode(ind) == dcode(ind-1),
            ccode = 2;     % straight
        elseif dcode(ind) - dcode(ind-1) == 1,
            ccode = 3;     % right
        elseif dcode(ind) ==1 && dcode(ind-1) == 4,
            ccode = 3;
        else
            ccode = 1;     % left
        end
    else
        ccode = 4;         % end-code
    end
    tempCcode(ind) = ccode;
    
    % update interval 
    b = b + sum(prob(1:ccode-1)) * l;
    l = prob(ccode) * l;
    
    % carry propagation
    if b>=1  % there is a carry
        b = b-1;
        d = Propagate_Carry(bCount,d);
    end
    
    % for debug ***
    tempBit = bCount;
    
    % interval renormalization
    while( l <= 0.5 )
        bCount = bCount+1;
        l = 2*l;
        if b >= 0.5
            d(bCount) = 1;  % output bit 1
            delta = 0.5;
            b = gamma*(b-delta);
        else
            d(bCount) = 0;  % output bit 0
            delta = 0;          
            b = gamma*(b-delta);
        end
    end  
    
    % for debug ***
    tempBit = bCount - tempBit;
    
    if tempMaxOrder ~= 0 && tempMaxOrder == maxOrder
        contextBitsOut(tempCurOrder, prohInd, 1) = 1 ...
            + contextBitsOut(tempCurOrder, prohInd, 1);
        
        contextBitsOut(tempCurOrder, prohInd, 2) = tempBit ...
            + contextBitsOut(tempCurOrder, prohInd, 2);
    end
    
end

% step 3: choose final code value ********
bCount = bCount+1;
if b <= 0.5
    d(bCount) = 1;
else
    d(bCount) = 0;
    d = Propagate_Carry(bCount-1,d);
end

% remove the 0's at the end
while(d(end)==0)
    d(end) = [];
    bCount = bCount-1;
end

end
